/*
 * Copyright (C) 2003 wwcube (wwcube@yahoo.co.jp)
 */

#include <sys/bios.h>

#include "4x4print.h"

static char far *fontbuf = (void*)0;

static int cursor_x = 0;
static int cursor_y = 0;

static int width = 40;
static int height = 20;
static int heightdot = 5;	/* 100/height */

int ww_4x4_locate(int x, int y)
{
	if ((x >= 0) && (x < width)){
		cursor_x = x;
	}
	if ((y >= 0) && (y < height)){
		cursor_y = y;
	}
	return 0;
}

int ww_4x4_initfont(char far *buf)
{
	fontbuf = MK_FP(FP_SEG(buf), FP_OFF(buf));
	return 0;
}

int ww_4x4_putc(int x, int y, int c)
{
	char far *p;
	c &= 0xff;
	p = fontbuf + c * 8;
	ww_put_4x4(x*4, y*heightdot, p);
	return 0;
}

int ww_4x4_putc_xor(int x, int y, int c)
{
	char far *p;
	c &= 0xff;
	p = fontbuf + c * 8;
	ww_put_4x4_xor(x*4, y*heightdot, p);
	return 0;
}

int ww_4x4_putcursor()
{
	ww_4x4_putc_xor(cursor_x, cursor_y, 0x87);
	return 0;
}

int ww_4x4_printchar(int c)
{
	int controlchar = 1;

	c &= 0xff;

	if (c == '\n'){
		cursor_x = 0;
		cursor_y++;
		if (cursor_y >= height){
			cursor_y = 0;
		}
	} else if (c == '\r'){
		cursor_x = 0;
	} else if (c == '\t'){
		cursor_x = (cursor_x + 8) & ~7;
		if (cursor_x >= width){
			cursor_x = 0;
			cursor_y++;
			if (cursor_y >= height){
				cursor_y = 0;
			}
		}
	} else {
		controlchar = 0;
		ww_4x4_putc(cursor_x, cursor_y, c);
		cursor_x++;
		if (cursor_x >= width){
			cursor_x = 0;
			cursor_y++;
			if (cursor_y >= height){
				cursor_y = 0;
			}
		}
	}
	return controlchar;
}

int ww_4x4_printstr(char far *str)
{
	char far *p;
	for (p = str; *p != '\0'; p++){
		ww_4x4_printchar((unsigned char)(*p));
	}
	return 0;
}

int ww_4x4_putchar(int c)
{
	ww_4x4_putc(cursor_x, cursor_y, c);
	cursor_x++;
	if (cursor_x >= width){
		cursor_x = 0;
		cursor_y++;
		if (cursor_y >= height){
			cursor_y = 0;
		}
	}
	return 0;
}

int ww_4x4_putstr(char far *str)
{
	char far *p;
	for (p = str; *p != '\0'; p++){
		ww_4x4_putchar((unsigned char)(*p));
	}
	return 0;
}
