/*
 * Copyright (C) 2003 wwcube (wwcube@yahoo.co.jp)
 */

#include <sys/bios.h>

#include <sys/libwwc.h>

#include <sys/types.h>
#include <sys/service.h>

#include <sys/process.h>

#include "wwgsub.h"
#include "wwkey.h"
#include "wwsound.h"

#include "4x4print.h"

#include "display.h"

#include "version.h"

#define DEMO_VOLUME	2
#define DEMO_HALF_SEC_WAIT	37

static int s_demo_skip = 0;

static int demo_force_wait(int n)
{
	sys_wait(n*DEMO_HALF_SEC_WAIT);
	return 0;
}

static int demo_wait(int n)
{
	if (!s_demo_skip) {
		int i;
		for (i = 0; i < n*DEMO_HALF_SEC_WAIT; i++){
			if (ww_get_keypress()){
				s_demo_skip = 1;
				break;
			}
			sys_wait(1);
		}
	}
	return 0;
}

static int demo_beep()
{
	if (!s_demo_skip) {
		int i;
		ww_stop_sound();
		ww_play_sound_volume(DEMO_VOLUME);
		for (i = 0; i < DEMO_HALF_SEC_WAIT; i++){
			if (ww_get_keypress()){
				s_demo_skip = 1;
				break;
			}
			sys_wait(1);
		}
		ww_stop_sound();
	}
	return 0;
}

static int demo_typestr(char far *str, int wait)
{
	char far *p;
	if (s_demo_skip) {
		wait = 0;
	}
	for (p = str; *p != '\0'; p++){
		int j;
		if (!ww_4x4_printchar((unsigned char)(*p))){
			ww_4x4_putcursor();
			for (j = 0; j < wait; j++){
				if (ww_get_keypress()){
					s_demo_skip = 1;
					wait = 0;
				}
				sys_wait(1);
			}
			ww_4x4_putcursor();
		}
	}
	return 0;
}

static int demo_blinkcursor(int numblink)
{
	int i;

	if (s_demo_skip) {
		return 1;
	}
	for (i = 0; i < numblink*2; i++){
		int j;
		ww_4x4_putcursor();
		for (j = 0; j < DEMO_HALF_SEC_WAIT; j++){
			if (ww_get_keypress()){
				if ((i & 1) == 0){
					ww_4x4_putcursor();
				}
				return 1;
			}
			sys_wait(1);
		}
	}
	return 0;
}

static void demo_draw_function()
{
	int x;

	/* t@NVL[ */
	ww_4x4_locate(2,19);
	ww_4x4_putstr("\t      auto   go to  list   run\15");
	for (x = 2; x <= 30; x += 7) {
		int i;
		for (i = 0; i < 6; i++) {
			ww_4x4_putc_xor(x+i, 19, '\207');
		}
	}
}

#define FADE_WAIT 5

static void set_monopalette(unsigned int i)
{
	unsigned int n;

	i &= 7;

	n = (i << 12) | (((i > 2) ? i - 2 : 0) << 8) | (((i > 4) ? i - 4 : 0) << 4);
	palette_set_color(FED_COLOR_SCREEN2, n);

	i = (i > 4) ? i - 4 : 0;
	n = (i << 12) | (((i > 2) ? i - 2 : 0) << 8) | (((i > 4) ? i - 4 : 0) << 4);
	palette_set_color(FED_COLOR_SCREEN1, n);
	palette_set_color(FED_COLOR_SCREEN1_DARK, n);
	palette_set_color(FED_COLOR_SCREEN2_DARK, n);
}

static void set_colorpalette(int i)
{
	unsigned int n;
	unsigned int c;
	wwc_palette_set_color(FED_COLOR_SCREEN2, 0, 0);
	c = (i > 7) ? i - 7 : 0;
	n = (c << 8)|(c<<4)|c;
	wwc_palette_set_color(FED_COLOR_SCREEN2, 1, n);
	c = (i > 3) ? i - 3 : 0;
	n = (c << 8)|(c<<4)|c;
	wwc_palette_set_color(FED_COLOR_SCREEN2, 2, n);
	n = (i << 8)|(i<<4)|i;
	wwc_palette_set_color(FED_COLOR_SCREEN2, 3, n);

	i = (i > 7) ? i - 7 : 0;

	wwc_palette_set_color(FED_COLOR_SCREEN1, 0, 0);
	wwc_palette_set_color(FED_COLOR_SCREEN1_DARK, 0, 0);
	wwc_palette_set_color(FED_COLOR_SCREEN2_DARK, 0, 0);
	c = (i > 7) ? i - 7 : 0;
	n = (c << 8)|(c<<4)|c;
	wwc_palette_set_color(FED_COLOR_SCREEN1, 1, n);
	wwc_palette_set_color(FED_COLOR_SCREEN1_DARK, 1, n);
	wwc_palette_set_color(FED_COLOR_SCREEN2_DARK, 1, n);
	c = (i > 3) ? i - 3 : 0;
	n = (c << 8)|(c<<4)|c;
	wwc_palette_set_color(FED_COLOR_SCREEN1, 2, n);
	wwc_palette_set_color(FED_COLOR_SCREEN1_DARK, 2, n);
	wwc_palette_set_color(FED_COLOR_SCREEN2_DARK, 2, n);
	n = (i << 8)|(i<<4)|i;
	wwc_palette_set_color(FED_COLOR_SCREEN1, 3, n);
	wwc_palette_set_color(FED_COLOR_SCREEN1_DARK, 3, n);
	wwc_palette_set_color(FED_COLOR_SCREEN2_DARK, 3, n);
}

static int demo_fade_in()
{
	if (wwc_get_hardarch() == HARDARCH_WS){
		int i;
		for (i = 0; i <= 7; i++) {
			set_monopalette(i);
			sys_wait(FADE_WAIT);
		}
	} else {
		int i;
		for (i = 1; i <= 15; i += 2) {
			set_colorpalette(i);
			sys_wait(FADE_WAIT);
		}
		set_colorpalette(15);
	}
}

static int demo_fade_out()
{
	if (wwc_get_hardarch() == HARDARCH_WS){
		int i;
		for (i = 7; i >= 0; i--) {
			set_monopalette(i);
			sys_wait(FADE_WAIT);
		}
	} else {
		int i;
		for (i = 15; i >= 0; i -= 2) {
			set_colorpalette(i);
			sys_wait(FADE_WAIT);
		}
		set_colorpalette(0);
	}
}

static int demo_wipe_out()
{
	int i;

	/* c */
	for (i = 0 ; i < 50; i += 3) {
		int j;
		ww_g_line(0,i, 159,i, 3);
		ww_g_line(0,100-i, 159,100-i, 3);
		sys_wait(1);
		ww_g_line(0,i, 159,i, 0);
		ww_g_line(0,100-i, 159,100-i, 0);
		ww_g_line(0,i+1, 159,i+1, 0);
		ww_g_line(0,99-i, 159,99-i, 0);
		ww_g_line(0,i+2, 159,i+2, 0);
		ww_g_line(0,98-i, 159,98-i, 0);
	}

	ww_g_line(0,50, 159,50, 3);

	/*  */
	for (i = 0; i < 80; i += 3) {
		ww_g_line(i,50, i+2,50, 0);
		ww_g_line(159-i,50, 157-i,50, 0);
		sys_wait(1);
	}

	ww_pset(80,50,3);

	demo_fade_out();

	ww_pset(80,50,0);
}

static void demo_init_palette(unsigned int i)
{
	if (wwc_get_hardarch() == HARDARCH_WS){
		lcd_set_color(0x7530, 0xfdb9);
		i = (i+1)/2;
		if (i > 1) i--;
		set_monopalette(i);
	} else {
		set_colorpalette(i);
	}
}

static void demo_init(int pal)
{
	s_demo_skip = 0;

	/* L[Ăԃ[v */
	ww_wait_keyrelease();

	/* ʃNA */
	fed_dsp_cls(0);

	/* TEh OFF */
	ww_stop_sound();

	/* pbg */
	demo_init_palette(pal);

}

static int demo_display_version()
{
	int i;

	/* o[W̕\ */
	demo_init(15);

	fed_dsp_set_contrast(3);

    fed_dsp_put_fedoralogo(16, 12, FED_COLOR_SCREEN2);
	ww_4x4_locate(14,3);
	ww_4x4_printstr("for WonderWitch V" FEDORA_VERSION);

	ww_4x4_locate(5,5);
	ww_4x4_printstr(
		"Copyright (C) 1983"
	);
	fed_dsp_put_yasuoka(96, 24, FED_COLOR_SCREEN2);

	ww_4x4_locate(6,8);
	ww_4x4_printstr("ported by wwcube@yahoo.co.jp");
	ww_4x4_locate(7,19);
	ww_4x4_printstr("BUILD: " __DATE__ " " __TIME__);

	for (i = 0; i < 750; i++) {
		if (ww_get_keypress() != 0) {
			break;
		}
		sys_wait(1);
	}

	demo_fade_out();

	fed_dsp_cls(0);
	fed_dsp_set_contrast(0);

	return 0;
}

/*
 * Q[JnÕf
 */
int game_start_demo()
{
	int i;

	/* Copyright & Version */
	demo_display_version();

	demo_init(0);

	/* t@NVL[ */
	demo_draw_function();

	/* NbZ[W */
	ww_4x4_locate(0,0);
	ww_4x4_printstr(
				"NEC PC-8001 BASIC Ver 1.0\n"
				"Copyright 1979 (C) by Microsoft\n"
				"\n"
				"Ok\n"
	);

	demo_fade_in();

	demo_blinkcursor(2);

	demo_typestr("clear 300,&hdbff\n",10);

	ww_4x4_printstr("Ok\n");

	demo_blinkcursor(1);

	demo_typestr("mon\n", 10);

	ww_4x4_printstr("*");

	demo_blinkcursor(1);

	demo_typestr("L\r", 10);

	demo_blinkcursor(1);

	if (!s_demo_skip) {
		ww_play_sound_volume(DEMO_VOLUME);
	}

	demo_blinkcursor(2);

	if (!s_demo_skip) {
		for (i = 0; i < 20; i++){
			int j;
			ww_4x4_putcursor();
			for (j = 0; j < DEMO_HALF_SEC_WAIT; j++){
				ww_change_sound_pitch(DEFAULT_PITCH * ((rand()&1) + 1));
				if (ww_get_keypress()){
					if ((i & 1) == 0){
						ww_4x4_putcursor();
					}
					i = 256;
					s_demo_skip = 1;
					break;
				}
				sys_wait(1);
			}
		}
		ww_change_sound_pitch(DEFAULT_PITCH);
	}

	demo_blinkcursor(1);

	ww_4x4_printstr("\n*");

	if (!s_demo_skip) {
		for (i = 0; i < 2; i++){
			int j;
			ww_4x4_putcursor();
			for (j = 0; j < DEMO_HALF_SEC_WAIT; j++){
				int d;
				d = i*DEMO_HALF_SEC_WAIT+j;
				if (d < 20){
					ww_change_sound_pitch(DEFAULT_PITCH+(i*DEMO_HALF_SEC_WAIT+j));
					ww_change_sound_volume(DEMO_VOLUME - (d/10));
				} else if (d == 20){
					ww_change_sound_volume(0);
				}
				if (ww_get_keypress()){
					if ((i & 1) == 0){
						ww_4x4_putcursor();
					}
					i = 256;
					s_demo_skip = 1;
					break;
				}
				sys_wait(1);
			}
		}
		ww_stop_sound();
	}

	demo_blinkcursor(1);

	demo_typestr("GDC00\n", 10);

	ww_4x4_printstr("FEDORA-2 Ready? ");

	demo_wait(4);

	ww_4x4_printstr("Yes\n");

	demo_beep();

	ww_4x4_printstr("     \225\225\n"
					"Push\227\210when you use it.\n"
					"     \224\224\n"
	);

	demo_wait(1);

	ww_4x4_printstr(" E000޽\n");

	demo_beep();

	ww_4x4_printstr("Ok\n");
	demo_blinkcursor(1);

	demo_typestr("mon\n", 10);

	ww_4x4_printstr("*");
	demo_blinkcursor(1);

	demo_typestr("GE000\n", 10);

	/* Ô߃TEh OFF */
	ww_stop_sound();

	/* L[Ăԃ[v */
	ww_wait_keyrelease();

	return 0;
}


/*
 * Q[Ĩf
 */
int game_end_demo()
{
	int key;
	int suspend = 0;

	demo_init(15);

	/* t@NVL[ */
	demo_draw_function();

	ww_4x4_locate(10,7);
	ww_4x4_printstr("CONTINUE: [A]    ");
	ww_4x4_locate(10,8);
	ww_4x4_printstr(" SUSPEND: [Y1]   ");
	ww_4x4_locate(10,9);
	ww_4x4_printstr("    EXIT: [START]");

	/* STOP Reset bZ[W */
	ww_4x4_locate(0,0);
	ww_4x4_printstr("Ok\n");

	for(;;) {
		s_demo_skip = 0;
		demo_blinkcursor(1);
		key = ww_get_keypress();
		if (key == KEY_START) {
			/* I */
			suspend = -1;
			break;
		} else if (key == KEY_Y1) {
			/* TXyh */
			suspend = 1;
			break;
		} else if (key == KEY_A) {
			/* ReBj[ */
			break;
		}
	}

	ww_4x4_locate(10,7);
	ww_4x4_printstr("                 ");
	ww_4x4_locate(10,8);
	ww_4x4_printstr("                 ");
	ww_4x4_locate(10,9);
	ww_4x4_printstr("                 ");

	if (suspend == -1) {
		/* I */
		demo_wipe_out();
		return 0;
	}

	if (suspend) {
		/* TXyh */
		ww_4x4_locate(16,8);
		ww_4x4_printstr("SUSPEND");

		ww_4x4_locate(0,1);
		demo_force_wait(3);

		demo_fade_out();

		/*
		 * TXyh/W[
		 */
		procIL->_yield();

		demo_init_palette(0);

		ww_4x4_locate(16,8);
		ww_4x4_printstr("RESUME ");

		demo_fade_in();

		/* L[Ăԃ[v */
		ww_wait_keyrelease();

		demo_force_wait(3);

		ww_4x4_locate(16,8);
		ww_4x4_printstr("       ");

		s_demo_skip = 0;
		ww_4x4_locate(0,1);
		demo_blinkcursor(2);
	} else {
		/* CONTINUE */
		ww_wait_keyrelease();
		s_demo_skip = 0;
	}
	ww_4x4_locate(0,1);
	demo_typestr("mon\n", 10);

	ww_4x4_printstr("*");
	demo_blinkcursor(1);

	demo_typestr("GE000\n", 10);

	/* L[Ăԃ[v */
	ww_wait_keyrelease();

	return 1;
}
