/*
 * Copyright (C) 2003 wwcube (wwcube@yahoo.co.jp)
 */

#include <sys/bios.h>

#include <sys/libwwc.h>

#include <sys/types.h>
#include <sys/service.h>
#include <sys/process.h>

#include <stdlib.h>

#include "wwkey.h"
#include "wwsound.h"
#include "wwsync.h"
#include "asmsub.h"

#include "display.h"
#include "4x4print.h"

#include "demo.h"

/* ======================================================================= */
/* ^`                                                                  */
/* ======================================================================= */

/* FEDORA ̏ԊǗ */
struct fedora_stat{
	unsigned char	f_x;
	unsigned char	f_y;
	char			f_xvector;
	char			f_yvector;
	unsigned char	f_xcounter;
	unsigned char	f_ycounter;
	unsigned char	f_direction;
	unsigned char	f_keyincounter;
};

/* TRILBY ̏ԊǗ */
struct trilby_stat{
	unsigned char	t_x;
	unsigned char	t_y;
	char			t_xvector;
	char			t_yvector;
	unsigned char	t_xcounter;
	unsigned char	t_ycounter;
	unsigned char	t_direction;
	unsigned char	t_keyincounter;
};

/* Cannon,j ̏ԊǗ */
struct cannon_stat{
	unsigned char	c_x;
	unsigned char	c_y;
	char			c_xvector;
	char			c_yvector;
	unsigned char	c_xcounter;
	unsigned char	c_ycounter;
	unsigned char	c_xcountbase;
	unsigned char	c_ycountbase;
};

/* ======================================================================= */
/* pf[^                                                          */
/* ======================================================================= */

/* FEDORA */
static struct fedora_stat fedora_init_data = {
	2,	/* XW */
	4,	/* YW */
	1,	/* XxNg */
	1,	/* YxNg */
	32,	/* XJE^ */
	39,	/* YJE^ */
	11,	/* ispx */
	32,	/* L[CJE^ */
};

/* TRILBY */
static struct trilby_stat trilby_init_data[] = {
	{
		0,	/* XW */
		0,	/* YW */
		-1,	/* XxNg */
		-1,	/* YxNg */
		32,	/* XJE^ */
		39,	/* YJE^ */
		27,	/* ispx */
		64,	/* L[CJE^ */
	},
	{
		0,	/* XW */
		0,	/* YW */
		1,	/* XxNg */
		1,	/* YxNg */
		32,	/* XJE^ */
		39,	/* YJE^ */
		11,	/* ispx */
		64,	/* L[CJE^ */
	},
	{
		0,	/* XW */
		0,	/* YW */
		1,	/* XxNg */
		-1,	/* YxNg */
		39,	/* XJE^ */
		32,	/* YJE^ */
		3,	/* ispx */
		64,	/* L[CJE^ */
	},
	{
		0,	/* XW */
		0,	/* YW */
		-1,	/* XxNg */
		1,	/* YxNg */
		39,	/* XJE^ */
		32,	/* YJE^ */
		19,	/* ispx */
		64,	/* L[CJE^ */
	},
};


/* ======================================================================= */
/* ֐vg^Cv                                                        */
/* ======================================================================= */
void atari_init();
int get_atari_charno(int no, int __ss *xp, int __ss *yp);

static int c_setwork(unsigned char x, unsigned char y, char xv, char yv, unsigned char xc, unsigned char yc);


#define NUMBEROF(s)	(sizeof(s)/sizeof(s[0]))

#define MAX_CANNON_STAT	32

/* ======================================================================= */
/* Q[pϐ                                                            */
/* ======================================================================= */
static unsigned char s_score;
static unsigned char s_soundcount;
static unsigned char s_beamcount;
static int s_beampower;
static unsigned char s_fedorawaitcount;
static unsigned char s_fedoracount;
static struct fedora_stat s_fedora;
static struct trilby_stat s_trilby[4];
static struct cannon_stat s_cannon[MAX_CANNON_STAT];

static unsigned char s_hiscore;

static int s_fedora2;

static int s_contrast = 1;

/* -------------------- */
/* ^Cg         */
/* -------------------- */
int game_title()
{
	int x;
	int y;
	int s;
	int i;

	/* LCD  */
	fed_dsp_off();

	/* XvCgȂ */
	sprite_set_range(0, 0); /* 擪ԍ,  */
	/* BMPSaver ΍ */
	fed_dsp_hide_fedorasprite();
	for (i = 0; i < NUMBEROF(s_trilby); i++) {
		fed_dsp_hide_trilbysprite(i);
	}
	for (i = 0; i < NUMBEROF(s_cannon); i++) {
		fed_dsp_hide_cannonsprite(i);
	}

	/*  */
	fed_dsp_cls();

	/* LCD CWP[^ */
	fed_dsp_set_contrast(s_contrast);

	/* L[ɐFĂ */
	for (y = 8; y < 13; y++) {
		for (x = 2; x < 8; x++) {
			fed_dsp_set_color(x,y, FED_COLOR_WHITE);
		}
	}
	for (y = 10; y < 11; y++) {
		for (x = 10; x < 20; x++) {
			fed_dsp_set_color(x,y, FED_COLOR_WHITE);
		}
	}

	/* "FEDORA" */
	fed_dsp_put_fedoralogo(4, 0, FED_COLOR_GREEN);

	/* "BY K.YASUOKA" */
	fed_dsp_put_yasuoka(48, 2, FED_COLOR_CYAN);

	/* "Hi-SCORE" */
	fed_dsp_put_hiscore(s_hiscore);

	/* "SCORE" */
	fed_dsp_put_score();

	/* LCD _ */
	fed_dsp_on();

	/* Score CTRILBY \(ő130) */
	x = 2;
	y = 22;
	s = s_score;
	if (s > 130) {
		s = 130;
	}
	for (i = 0; i < s; i++) {
		if (s_hiscore <= i) {
			s_hiscore = i+1;
			fed_dsp_put_num3(90, 12, s_hiscore);
		}
		fed_dsp_put_num3(86, 60, i+1);
		fed_dsp_put_fedora(x, y);
		x += 6;
		if ((i % 26) == 25) {
			x = 2;
			y += 8;
		}
		ww_play_sound();
		sys_wait(3);
		ww_stop_sound();
		sys_wait(3);
	}

	/* L[̕\ */
	if (s_fedora2) {
		/* FEDORA-2 p\ */
		fed_dsp_put_keydesc2();
	} else {
		/* FEDORA p\ */
		fed_dsp_put_keydesc();
	}
	for (;;) {
		int key;
		key = ww_get_keypress();
		if (key & KEY_START) {
			/* return to OS */
			return -1;
		} else if (key & KEY_Y1) {
			if (s_fedora2) {
				/* FEDORA-2 p\ */
				fed_dsp_put_keydesc2();
				fed_dsp_put_keydesc();
			} else {
				/* FEDORA p\ */
				fed_dsp_put_keydesc();
				fed_dsp_put_keydesc2();
			}
			s_fedora2 ^= 1;
			while ((key = ww_get_keypress()) & KEY_Y1) {
				if (key & KEY_START) {
					return -1;
				}
				sys_wait(1);
			}
		} else if (key & (KEY_A|KEY_B|KEY_Y3)) {
			break;
		} else if (key & (KEY_Y4|KEY_Y2)) {
			s_contrast++;
			s_contrast &= 3;
			fed_dsp_set_contrast(s_contrast);
			ww_wait_keyrelease();
		}
		sys_wait(1);
	}

	return s_contrast;
}

/* -------------------------- */
/* Q[pϐ         */
/* -------------------------- */
void game_var_init()
{
	static struct cannon_stat cannon_init_data = {0,0,0,0,0,0,0,0};
	int i;
	s_score = 0;
	s_soundcount = 0;
	s_beamcount = 0;
	s_beampower = 0;
	s_fedorawaitcount = 0;
	s_fedoracount = 2;
	s_fedora = fedora_init_data;
	for (i = 0; i < NUMBEROF(s_trilby); i++) {
		s_trilby[i] = trilby_init_data[i];
	}
	for (i = 0; i < NUMBEROF(s_cannon); i++) {
		s_cannon[i] = cannon_init_data;
	}
}

/* -------------------------- */
/* Q[pʏ         */
/* -------------------------- */
void game_screen_init()
{
	fed_dsp_put_game_screen();
}

int trilby_cannon(struct trilby_stat far *trilbyp, int angle)
{
	int kc;
	int xv;
	int xc;
	int yv;
	int yc;

	if (s_soundcount <= 24) {
		s_soundcount = 24;
	}

	angle2vc(angle, &xv, &xc);
	angle2vc(angle-8, &yv, &yc);

	xc >>= 1;
	yc >>= 1;

	/* cannon ˈʒuvZ */
	kc = trilbyp->t_keyincounter;
	c_startpos(&kc, angle, trilbyp->t_x, trilbyp->t_y, xv, yv, xc, yc);
	trilbyp->t_keyincounter = kc;

	return 0;
}

/* ------------------------------------ */
/* xNgvZCTRILBYCannon˔f */
/* ------------------------------------ */
int trilby_vec_cannon(struct trilby_stat far *trilbyp)
{
	static int vtable[] = {
		0x03, 0x06, 0x0A, 0x10, 0x17, 0x26, 0x50,
	};
	int i;
	int dx;
	int dy;
	int vc;
	int vd;
	int t_dir;
	int f_dir;

	trilbyp->t_keyincounter--;
	if (trilbyp->t_keyincounter != 0) {
		return 0;
	}
	trilbyp->t_keyincounter = 64;

	vc = 8;
	vd = 1;

	dx = (int)(s_fedora.f_x) - (int)(trilbyp->t_x);
	if (dx < 0) {
		dx = -dx;
		vc = 23;
		vd = -1;
	}

	dy = (int)(s_fedora.f_y) - (int)(trilbyp->t_y);
	if (dy < 0) {
		dy = -dy;
		vc = (47 - vc) & 0x1f;
		vd = -vd;
	}

	for (i = 0; i < NUMBEROF(vtable); i++) {
		if ((dx * vtable[i]) >= (dy * 16)) {
			break;
		}
		vc += vd;
	}

	t_dir = (trilbyp->t_direction) & 0x1f;
	f_dir = (s_fedora.f_direction) & 0x1f;

	if (f_dir != t_dir) {
		/* Ε */
		if (f_dir < t_dir) {
			f_dir += 16;
		} else {
			f_dir -= 16;
		}
		f_dir += t_dir;
		f_dir = (f_dir >> 1) & 0x1f;
	}

	f_dir -= vc;
	if (((vc - f_dir) & 0x1f) == t_dir) {
		trilby_cannon(trilbyp, t_dir);
	}

	if (f_dir & 0x10) {
		(trilbyp->t_direction)++;
	} else {
		(trilbyp->t_direction)--;
	}

	return 0;
}

/* -------------------------- */
/* px  xNgCJE^ */
/* -------------------------- */
int angle2vc(int angle, int __ss *vectorp, int __ss *counterp)
{
	static int angle2vec_table[] = {
			85, 29, 18, 13, 11,  9,  9,  8,
			 8,  9,  9, 11, 13, 18, 29, 85,
	};

	if (angle & 0x10) {
		*vectorp = -1;
	} else {
		*vectorp = 1;
	}
	*counterp = angle2vec_table[angle & 0x0F];
	return 0;
}

/* -------------------------- */
/* TRILBY isω        */
/* -------------------------- */
int trilby_move()
{
	int i;

	if (s_fedora.f_x == 0) {
		return 0;
	}

	for (i = 0; i < NUMBEROF(s_trilby); i++) {
		if (s_trilby[i].t_x != 0) {
			trilby_vec_cannon(&s_trilby[i]);
		}
	}

	return 0;
}

/* -------------------------- */
/* FEDORA,TRILBY Xړ  */
/* -------------------------- */
int fedora_trilby_x_move()
{
	int i;

	/* FEDORA */
	if (s_fedora.f_x != 0) {
		s_fedora.f_xcounter--;
		if (s_fedora.f_xcounter == 0) {
			int v;
			int c;
			int newx;
			int x;
			int y;
			angle2vc(s_fedora.f_direction, &v, &c);
			s_fedora.f_xvector = v;
			s_fedora.f_xcounter = c;
			/* ł͂ňULN^ */
			newx = s_fedora.f_x + s_fedora.f_xvector;
			if ((newx == 1) || (newx == 154)) {
				s_fedora.f_direction = 31 - s_fedora.f_direction;
				newx = s_fedora.f_x - s_fedora.f_xvector;
			}
			s_fedora.f_x  = newx;
			/* 蔻 & LN^\ */
			if (get_atari_charno(0, &x, &y) != -1) {
				/* Փ */
				clash(0, x, y);
			} else {
				fed_dsp_move_fedorasprite(s_fedora.f_x, s_fedora.f_y);
			}
		}
	}

	/* TRILBY */
	for (i = 0; i < NUMBEROF(s_trilby); i++) {
		if (s_trilby[i].t_x != 0) {
			s_trilby[i].t_xcounter--;
			if (s_trilby[i].t_xcounter == 0) {
				int v;
				int c;
				int newx;
				int x;
				int y;
				angle2vc(s_trilby[i].t_direction, &v, &c);
				s_trilby[i].t_xvector = v;
				s_trilby[i].t_xcounter = c;
				/* ł͂ňULN^ */
/*
				sprite_set_location(i+1, 224, 224);
*/
				newx = s_trilby[i].t_x + s_trilby[i].t_xvector;
				if ((newx == 1) || (newx == 154)) {
					s_trilby[i].t_direction = 31 - s_trilby[i].t_direction;
					newx = s_trilby[i].t_x - s_trilby[i].t_xvector;
				}
				s_trilby[i].t_x  = newx;
				/* 蔻 & LN^\ */
				if (get_atari_charno(i+1, &x, &y) != -1) {
					/* Փ */
					clash(i+1, x, y);
				} else {
					/* LN^\ */
					fed_dsp_move_trilbysprite(i, s_trilby[i].t_x, s_trilby[i].t_y);
				}
			}
		}
	}

	return 0;
}

/* -------------------------- */
/* FEDORA,TRILBY Yړ  */
/* -------------------------- */
int fedora_trilby_y_move()
{
	int i;

	/* FEDORA */
	if (s_fedora.f_x != 0) {
		s_fedora.f_ycounter--;
		if (s_fedora.f_ycounter == 0) {
			int v;
			int c;
			int newy;
			int x;
			int y;
			angle2vc(s_fedora.f_direction - 8, &v, &c);
			s_fedora.f_yvector = v;
			s_fedora.f_ycounter = c;
			/* ł͂ňULN^ */
			newy = s_fedora.f_y + s_fedora.f_yvector;
			if ((newy == 3) || (newy == 92)) {
				s_fedora.f_direction = 47 - s_fedora.f_direction;
				newy = s_fedora.f_y - s_fedora.f_yvector;
			}
			s_fedora.f_y  = newy;
			/* 蔻 & LN^\ */
			if (get_atari_charno(0, &x, &y) != -1) {
				/* Փ */
				clash(0, x, y);
			} else {
				fed_dsp_move_fedorasprite(s_fedora.f_x, s_fedora.f_y);
			}
		}
	}

	/* TRILBY */
	for (i = 0; i < NUMBEROF(s_trilby); i++) {
		if (s_trilby[i].t_x != 0) {
			s_trilby[i].t_ycounter--;
			if (s_trilby[i].t_ycounter == 0) {
				int v;
				int c;
				int newy;
				int x;
				int y;
				angle2vc(s_trilby[i].t_direction - 8, &v, &c);
				s_trilby[i].t_yvector = v;
				s_trilby[i].t_ycounter = c;
				/* ł͂ňULN^ */
				newy = s_trilby[i].t_y + s_trilby[i].t_yvector;
				if ((newy == 3) || (newy == 92)) {
					s_trilby[i].t_direction = 47 - s_trilby[i].t_direction;
					newy = s_trilby[i].t_y - s_trilby[i].t_yvector;
				}
				s_trilby[i].t_y  = newy;
				/* 蔻 & LN^\ */
				if (get_atari_charno(i+1, &x, &y) != -1) {
					/* Փ */
					clash(i+1, x, y);
				} else {
					/* LN^\ */
					fed_dsp_move_trilbysprite(i, s_trilby[i].t_x, s_trilby[i].t_y);
				}
			}
		}
	}

	return 0;
}

/*
 * {ł́ChbgPʂ̃`FbNsĂ̂
 * LN^p^[ӎKvB
     *
 *    0+1+2+3+4
 *  0 . . o . .
 * +1 . o . o .
 * +2 o . . . o
 * +3 . o o o .
 * +4 o . o . o
 *
 */
static	struct r_xy {
	int xd;
	int yd;
} r_xy_table[] = {
					{2,0},
				{1,1},	{3,1},
			{0,2},			{4,2},
				{1,3},{2,3},{3,3},
			{0,4},	{2,4},	{4,4}
};

/*
  0 1 2 3 4 5 6 7 8
0 . . . . . . . . .
1 . . . . . . . . .
2 . . . . . . . . .
3 . . . . . . . . .
4 . . . . . . o . .
5 . . . . . o . o .
6 . . . . o . . . o
7 . . . . . o o o .
8 . . . . o . o . o
 */
static char atari[9][9];	/* (X size x 2) + 1, (Y size x 2) + 1 */

/*
 *  (4,4) ɂꍇC(x,y) ɂ铯^LN^ƏՓ˂邩H
 * Փ˂ȂCLN^\hbg̉hbgڂՓ˂̂H
 */
void atari_init()
{
	int x;
	int y;
	for (y = 0; y < 9; y++) {
		for (x = 0; x < 9; x++) {
			int f;
			int i;
			f = 0;
			/*  (4,4) ̃LN^\邷ׂẴhbgɂ */
			for (i = 0; (f == 0) && (i < NUMBEROF(r_xy_table)); i++) {
				int j;
				/* (x,y) ̃LN^\ǂ̃hbgɓ邩H */
				for (j = 0; (f == 0) && (j < NUMBEROF(r_xy_table)); j++) {
					if ((4 + r_xy_table[i].xd == x + r_xy_table[j].xd)
					 && (4 + r_xy_table[i].yd == y + r_xy_table[j].yd)) {
						/*  i hbgڂdȂ */
						f = i + 1;
					}
				}
			}
			atari[x][y] = f;
		}
	}
	return;
}

/* ----------------------------------------------------------------------------
 * w肵LN^Ƒ̃LN^Փ˂Ă邩H
 *
 * OUT
 *   *xp ... YLN^̏Փ˕ X W(ԉE)
 *   *yp ... YLN^̏Փ˕ Y W(ԉE)
 * -----------------------------------------------------------------------------
 */
int get_atari_charno(int no, int __ss *xp, int __ss *yp)
{
	int x;
	int y;
	if (no == 0) {
		/*  FEDORA */
		int i;
		x = s_fedora.f_x;
		y = s_fedora.f_y;
		for (i = 0; i < NUMBEROF(s_trilby); i++) {
			if (s_trilby[i].t_x != 0) {
				int dx;
				int dy;
				dx = s_trilby[i].t_x - x;
				dy = s_trilby[i].t_y - y;
				if (((dx < 5) && (dx > -5)) && ((dy < 5) && (dy > -5))) {
					int n;
					n = atari[dx+4][dy+4];
					if (n) {
						/* FEDORA  TRILBY ƏՓ */
						*xp = x + r_xy_table[n-1].xd;
						*yp = y + r_xy_table[n-1].yd;
						return i+1; /* TRILBY ԍԂ */
					}
				}
			}
		}
	} else {
		/*  TRILBY */
		int dx;
		int dy;
		int i;
		x = s_trilby[no-1].t_x;
		y = s_trilby[no-1].t_y;
		if (s_fedora.f_x != 0) {
			dx = s_fedora.f_x - x;
			dy = s_fedora.f_y - y;
			if ((dx < 5 && dx > -5) && (dy < 5 && dy > -5)) {
				int n;
				n = atari[dx+4][dy+4];
				if (n) {
					/* TRILBY  FEDORA ƏՓ */
					*xp = x + r_xy_table[n-1].xd;
					*yp = y + r_xy_table[n-1].yd;
					return 0; /* FEDORA Ԃ */
				}
			}
		}
		for (i = 0; i < NUMBEROF(s_trilby); i++) {
			if (i == (no-1)) {
				/* Ƃ̏Փ˂̓`FbNȂ */
				continue;
			}
			if (s_trilby[i].t_x != 0) {
				dx = s_trilby[i].t_x - x;
				dy = s_trilby[i].t_y - y;
				if (((dx < 5) && (dx > -5)) && ((dy < 5) && (dy > -5))) {
					int n;
					n = atari[dx+4][dy+4];
					if (n) {
						/* TRILBY  TRILBY ƏՓ */
						*xp = x + r_xy_table[n-1].xd;
						*yp = y + r_xy_table[n-1].yd;
						return i+1; /* TRILBY ԍԂ */
					}
				}
			}
		}
	}
	return -1;
}


/*
 * w肵W̓_ɁC邩Ԃ
 *
 *
 * -1 ... ȂɂȂ
 *  0 ... FEDORA
 *  1 ... TRILBY[0]
 *  2 ... TRILBY[1]
 *  3 ... TRILBY[2]
 *  4 ... TRILBY[3]
 */
int get_clash_no(int cx, int cy)
{
	int i;

	/*
	 * {ł́ChbgPʂ̃`FbNsĂ̂
	 * LN^p^[ӎKvB
     *
	 *    0+1+2+3+4
	 *  0 . . o . .
	 * +1 . o . o .
	 * +2 o . . . o
	 * +3 . o o o .
	 * +4 o . o . o
	 *
	 */
	if (s_fedora.f_x != 0) {	/* H */
		int x;
		int y;
		int j;
		x = s_fedora.f_x;
		y = s_fedora.f_y;
		for (j = 0; j < NUMBEROF(r_xy_table); j++) {
			if (((x + r_xy_table[j].xd) == cx) && ((y + r_xy_table[j].yd) == cy)) {
				return 0;	/* FEDORA clash! */
			}
		}
	}

	for (i = 0; i < NUMBEROF(s_trilby); i++) {
		if (s_trilby[i].t_x != 0) {	/* H */
			int x;
			int y;
			int j;
			x = s_trilby[i].t_x;
			y = s_trilby[i].t_y;
			for (j = 0; j < NUMBEROF(r_xy_table); j++) {
				if (((x + r_xy_table[j].xd) == cx) && ((y + r_xy_table[j].yd) == cy)) {
					return i+1;	/* TRILBY clash! */
				}
			}
		}
	}

	return -1;
}

int clash(int no, int x, int y)
{
	struct clash_vc {
		int xc;
		int yc;
		int xv;
		int yv;
	};
	static struct clash_vc vc[] = {
		{22,  6,  1, -1},
		{22,  6, -1,  1},
		{ 6, 22,  1,  1},
		{ 6, 22, -1, -1},
		{ 7, 13,  1, -1},
		{ 7, 13, -1,  1},
		{13,  7,  1,  1},
		{13,  7, -1, -1},
	};
	int i;

	if (no == 0) {
		/* LN^\ */
		fed_dsp_hide_fedorasprite();
		s_fedora.f_x = 0;
	} else {
		/* XRA{P */
		s_score++;
		/* LN^\ */
		fed_dsp_hide_trilbysprite(no-1);
		s_trilby[no-1].t_x = 0;
	}

	/* jЃ[NGAZbg */
	for (i = 0; i < NUMBEROF(vc); i++) {
		c_setwork(x, y, vc[i].xv, vc[i].yv, vc[i].xc, vc[i].yc);
	}

	s_soundcount = 255;

	return 0;
}

int cannon_erase(int n)
{
	fed_dsp_hide_cannonsprite(n);
	return 0;
}

int cannon_put(int n, int x, int y)
{
	int no;

	if ((x < 2) || (x >= 158) || (y < 4) || (y >= 96)) {
		return -1;
	}

	no = get_clash_no(x, y);
	if (no >= 0) {
		/* Փˏ */
		clash(no, x, y);
	}

	fed_dsp_move_cannonsprite(n, x, y);

	return 0;
}

int cannon_x_move()
{
	int i;

	for (i = 0; i < NUMBEROF(s_cannon); i++) {
		if (s_cannon[i].c_x != 0) {
			s_cannon[i].c_xcounter--;
			if (s_cannon[i].c_xcounter == 0) {
				s_cannon[i].c_x += s_cannon[i].c_xvector;
				if (cannon_put(i, s_cannon[i].c_x, s_cannon[i].c_y) == -1) {
					/* ʂ͂ݏo */
					cannon_erase(i);
					s_cannon[i].c_x = 0;
				}
				s_cannon[i].c_xcounter = s_cannon[i].c_xcountbase;
			}
		}
	}

	return 0;
}

int cannon_y_move()
{
	int i;
	for (i = 0; i < NUMBEROF(s_cannon); i++) {
		if (s_cannon[i].c_x != 0) {
			s_cannon[i].c_ycounter--;
			if (s_cannon[i].c_ycounter == 0) {
				s_cannon[i].c_y += s_cannon[i].c_yvector;
				if (cannon_put(i, s_cannon[i].c_x, s_cannon[i].c_y) == -1) {
					/* ʂ͂ݏo */
					cannon_erase(i);
					s_cannon[i].c_x = 0;
				}
				s_cannon[i].c_ycounter = s_cannon[i].c_ycountbase;
			}
		}
	}

	return 0;
}

static int c_setwork(unsigned char x, unsigned char y, char xv, char yv, unsigned char xc, unsigned char yc)
{
	int i;

	for (i = 0; i < NUMBEROF(s_cannon); i++) {
		if (s_cannon[i].c_x == 0) {
			/* [NGAݒ */
			s_cannon[i].c_x = x;
			s_cannon[i].c_y = y;
			s_cannon[i].c_xvector = xv;
			s_cannon[i].c_yvector = yv;
			s_cannon[i].c_xcounter = s_cannon[i].c_xcountbase = xc;
			s_cannon[i].c_ycounter = s_cannon[i].c_ycountbase = yc;

			/* \(Փ˃`FbNȂ) */
			(void)fed_dsp_move_cannonsprite(i, x, y);

			return 0;
		}
	}

	return -1;
}

int c_startpos(int __ss *kcp, int angle, int x, int y, int xv, int yv, int xc, int yc)
{
	struct c_relative_pos {
		int x;
		int y;
	};
	static struct c_relative_pos pos[] = {
		{2,-2},{3,-2},{4,-2},{4,-1},{5,-1},{5,0},{6,0},{6,1},
		{6,2},{6,3},{6,4},{6,5},{6,6},{5,6},{4,6},{3,6},
		{2,6},{1,6},{0,6},{-1,6},{-2,6},{-2,5},{-2,4},{-2,3},
		{-2,2},{-2,1},{-2,0},{-1,0},{-1,-1},{0,-1},{0,-2},{1,-2},
	};
	/*
		           310 1 2
		       30+ + + + +
		     28+9+       +3+4
		   26+7+ @         +5+6
		   25+               +7
		   24+               +8
		   23+               +9
		   22+               +10
		   21+               +11
		   20+ + + + + + + + +12
		       19181716151413
	*/

	/* L[CJE^ */
	*kcp |= 0x04;

	/* px 0`31  */
	angle &= 0x1f;

	/* ˈʒuvZ */
	x += pos[angle].x;
	if (x == 0) {
		return -1;
	}
	y += pos[angle].y;

	/*  */
	c_setwork(x, y, xv, yv, xc, yc);

	return 0;
}

int cannon(struct fedora_stat far *fsp, int angle)
{
	int kc;
	int xv;
	int xc;
	int yv;
	int yc;

	if (s_soundcount <= 48) {
		s_soundcount = 48;
	}

	s_beampower--;

	angle2vc(angle, &xv, &xc);
	angle2vc(angle-8, &yv, &yc);

	xc >>= 2;
	yc >>= 2;

	/* cannon ˈʒuvZ */
	kc = fsp->f_keyincounter;
	c_startpos(&kc, angle, fsp->f_x, fsp->f_y, xv, yv, xc, yc);
	fsp->f_keyincounter = kc;

	return 0;
}

int slow_cannon(struct fedora_stat far *fsp, int angle)
{
	int kc;
	int xv;
	int xc;
	int yv;
	int yc;

	if (s_soundcount <= 24) {
		s_soundcount = 24;
	}

	angle2vc(angle, &xv, &xc);
	angle2vc(angle-8, &yv, &yc);

	xc >>= 1;
	yc >>= 1;

	/* cannon ˈʒuvZ */
	kc = fsp->f_keyincounter;
	c_startpos(&kc, angle, fsp->f_x, fsp->f_y, xv, yv, xc, yc);
	fsp->f_keyincounter = kc;

	return 0;
}

int cannon_power_up()
{
	s_beamcount--;
	if (s_beamcount == 0) {
		if (s_beampower < 255) {
			s_beampower++;
		}
	}
	return 0;
}

int keyin(struct fedora_stat far *fsp)
{
	int key;
	int dir;

	/* L[ */
	key = ww_get_keypress();

	/* L[̔ */
	if (key & KEY_START) {
		return -1;
	}

	/* oH */
	if (fsp->f_x == 0) {
		return 0;
	}

	/* L[CJE^ */
	fsp->f_keyincounter--;
	if (fsp->f_keyincounter != 0) {
		return 0;
	}
	fsp->f_keyincounter = 32;

	if (s_fedora2) {
		/*
		 * FEDORA-2
		 */
		if (key & (KEY_X1|KEY_X2|KEY_X3|KEY_X4)) {
			dir = fsp->f_direction;
			if (key & KEY_X1) {
				dir -= 0;
			} else if (key & KEY_X3) {
				dir -= 16;
			} else if (key & KEY_X4) {
				dir -= 24;
			} else if (key & KEY_X2) {
				dir -= 8;
			}
			if ((dir & 0x10) == 0) {
				fsp->f_direction--;
			} else {
				fsp->f_direction++;
			}
		}

		/* r[o */
		if (s_beampower < 2) {
			return 0;
		}

		/* ݂̐is */
		dir = fsp->f_direction;

		/* O cannon */
		if (key & KEY_A) {
			cannon(fsp, dir);
		}
		/*  cannon */
		if (key & KEY_B) {
			s_beampower--;
			slow_cannon(fsp, dir + 16);
		}
	} else {
		/*
		 * FEDORA
		 */

		/*  */
		if (key & KEY_B) {
			/* ispx -1 */
			fsp->f_direction--;
		}
		/* E */
		if (key & KEY_A) {
			/* ispx +1 */
			fsp->f_direction++;
		}

		/* r[o */
		if (s_beampower < 6) {
			return 0;
		}

		/* ݂̐is */
		dir = fsp->f_direction;

		 /*  */
		if (key == (KEY_A | KEY_B | KEY_X1 | KEY_X2 | KEY_X3 | KEY_X4)) {
			key = (KEY_Y4 | KEY_X4 | KEY_X1 | KEY_X2 | KEY_Y2 | KEY_X3);
		}

		/* O */
		if (key & KEY_Y4) {
			cannon(fsp, dir - 4);
		}
		if (key & KEY_X4) {
			cannon(fsp, dir - 2);
		}
		if (key & KEY_X1) {
			cannon(fsp, dir);
		}
		if (key & KEY_X2) {
			cannon(fsp, dir + 2);
		}
		if (key & KEY_Y2) {
			cannon(fsp, dir + 4);
		}

		/*  */
		if (key & KEY_X3) {
			s_beampower--;
			slow_cannon(fsp, dir + 16);
		}
	}

	return 0;
}

int game_over(int winning)
{
	int n;

	if ((s_beamcount & 0x07) != 0) {
		return 0;
	}

	n = s_fedorawaitcount;

	s_fedorawaitcount--;
	if (s_fedorawaitcount == 0) {
		/* GAME OVER */
		return -1;
	}

	if ((n & 0x2f) != 0) {
		return 0;
	}

	/* "GAME OVER" or "WINNING" \,  */
	if (winning) {
		fed_dsp_put_winning();
	} else {
		fed_dsp_put_gameover();
	}

	return 0;
}

int shutsugen_check()
{
	/* o */
	if (s_soundcount != 0) {
		s_soundcount--;
		/* o */
		/*
		 *(1) BEEP ON (1200Hz)
		 *(2) Wait 0.000222856[s]
		 *(3) BEEP OFF
		 */
		ww_play_sound();
		spin_wait();
		ww_stop_sound();
		return 0;
	}

	/* GAME OVER `FbN */

	if (s_fedora.f_x == 0) {
		if (s_fedoracount == 0) {
			if (game_over(0)) {
				return 1;
			}
			return 0;
		}
		if ((s_beamcount & 0x07) == 0) {
			int key;
			key = ww_get_keypress();
			if (key & (KEY_Y1 | KEY_Y3)) {
				s_fedorawaitcount = 0;
			} else {
				s_fedorawaitcount--; /* o҂JE^-1 */
			}
			if (s_fedorawaitcount == 0) {
				s_fedoracount--;
				/* FEDORA o */
				s_fedora.f_x = 2;
				s_fedora.f_y = 4;

				s_beampower += 16;
				if (s_beampower > 255) {
					s_beampower = 255;
				}
			}
		}
	}

	/* Winning ? */
	if (s_score >= 130) {
		/* Winning */
		if (s_fedora.f_x == 0) {
			return 0;
		}
		if (game_over(1)) {
			return 1;
		}
	} else {
		/* TRILBY o */
		int a;
		int b;
		int i;
		a = s_score;
		b = 1;
		if ((a & 1) == 0) {
			if ((a & 2) == 0) {
				a = (a >> 3) & 3;
				a++;
				b = a;
			}
		}
		for (i = 0; i < b; i++) {
			if (s_trilby[i].t_x == 0) {
				struct xypos {
					unsigned char x;
					unsigned char y;
				};
				static struct xypos trilby_xy[] = {
					{153,  91}, /* TRILBY 1 */
					{  2,   4}, /* TRILBY 2 */
					{  2,  91}, /* TRILBY 3 */
					{153,   4}, /* TRYLBY 4 */
				};
				s_trilby[i].t_x = trilby_xy[i].x;
				s_trilby[i].t_y = trilby_xy[i].y;
				s_beampower += 16;
				if (s_beampower > 255) {
					s_beampower = 255;
				}
			}
		}
	}

	return 0;
}

void game_wait()
{
	int w;
	int i;
	int n;

	w = 0;
	for (i = 0; i < NUMBEROF(s_cannon); i++) {
		if (s_cannon[i].c_x != 0) {
			w += 5;
		}
	}
	w /= 3;

	n = 0;
	for (i = 0; i < NUMBEROF(s_trilby); i++) {
		if (s_trilby[i].t_x != 0) {
			w += 13;
			n++;
		}
	}

	if (s_fedora.f_x != 0) {
		w += (10 + n * 3);
	}

	w += 30;

	if (w > 511) {
		w = 511;
	}

	while (ww_sync_get_counter() < w) {
		;	/* DO NOTHING */
	}
	ww_sync_set_counter(0);
}

void game_main()
{
	s_score = 0;

	for (;;) {
		int rc;
		int contrast = 0;

		rc = game_title();
		if (rc == -1) {
			return;
		} else if (rc == 1) {
			contrast = 1;
		}

		game_var_init();
		game_screen_init();

		for (;;) {
			if (keyin(&s_fedora) < 0) {
				return;
			}
			trilby_move();
			fedora_trilby_x_move();
			fedora_trilby_y_move();
			cannon_x_move();
			cannon_y_move();
			cannon_power_up();
			if (shutsugen_check() != 0) {
				break;
			}
			game_wait();
		}
	}

	return;
}

void main(int argc, char *argv[])
{
	/* LN^m̓蔻pe[u */
	atari_init();

	/* 荞ݏ̏ */
	(void)ww_init_key();
	(void)ww_sync_init();

	/* L[Ăԃ[v */
	ww_wait_keyrelease();

	/* \[X̓ǂݏo */
	(void)fed_dsp_load_resources();

	/* ʂƉ */
	(void)fed_dsp_init_screen();
	(void)ww_init_sound();

	/* [h */
	game_start_demo();

	while(1) {
		/* ʂƉ */
		(void)fed_dsp_init_screen();
		(void)ww_init_sound();

		/* C */
		game_main();

		/* KvɉĉʃLv` */
		if (ww_get_keypress() & KEY_Y1) {
			fed_dsp_bmpsaver();
		}

		/* ʂ */
		(void)fed_dsp_init_screen();

		/* If */
		if (game_end_demo() == 0) {
			/* Exit */
			break;
		}

		/* Continue (Restart Game) */
	}
}
